/*
**  File:         endian.c
**  Purpose:      Read/write file stream data in big or 
**                     little-endian order.
**  Author:       James D. Murray, Anaheim, CA  USA
**  Date:         February 1990
**  C Compilers:  Any      
**  Memory Model: Any
**  
**  (c) 1990 Copyright The C Gazette. May be used freely if
**                  author and publication are acknowledged.
**
**  Parameter values: LSB_FIRST  read/write little-endian data
**                    MSB_FIRST  read/write big-endian data
**
**  The following five functions allow a program to test the
**  native byte order of a machine and to read and write data to
**  a file stream in either big-endian or little-endian order.
**  These I/O routines use fgetc() and fputc() along with simple
**  masking and shifting to read and write 16-bit and 32-bit
**  values in the indicated order.
**
**  Example code:
**
**  long i;                                                           
**  FILE *fp_in, *fp_out;                                             
**
**  /* Test native byte order */                                    
**  printf("This machine is %s-endian.\n",                            
**      TestByteOrder() == MSB_FIRST ? "big" : "little");             
**
**  i = fgetlong(MSB_FIRST, fp_in); /* Read as big-endian */        
**  fputlong(LSB_FIRST, fp_out, i); /* Write as little-endian */    
*/

#include <stdio.h>

#define LSB_FIRST       0   /* Little-endian */
#define MSB_FIRST       1   /* Big-endian */

/*
** Function prototypes
*/
int   TestByteOrder(void);
short fgetshort(char, FILE *);
void  fputshort(char, FILE *, int);
long  fgetlong(char, FILE *);
void  fputlong(char, FILE *, long);


/*
**  Test the native byte order.
**
**  An integer value is assigned to the upper and lower bytes
**  within a two-byte word. The word is then divided into
**  a high byte and a low byte. If the low byte is non-zero
**  then the machine is little-endian. If the low-byte is zero
**  then the machine is big-endian.
**
**  Little endian systems include the 80x86 family, 6502, and VAX
**  Big endian systems include the MC680x0 family and IBM 370 
**
**  Returns: LSB_FIRST if the machine is little-endian, otherwise
**           MSB_FIRST if the machine is big-endian.
*/

int TestByteOrder()
{
        short int word = 0x0001;
        char *byte = (char *) &word;

        return(byte[0] ? LSB_FIRST : MSB_FIRST);
}

/*
**  Read a long word (32-bits) from a file stream.
**
**  Parameter values: byteorder : LSB_FIRST or MSB_FIRST
**                    fp        : FILE stream
**  Returns: A 32-bit value in the specified format.
*/
long fgetlong(char byteorder, FILE *fp)
{
        register long word;

        if (byteorder == MSB_FIRST) {
            word =  (long)  fgetc(fp) & 0xff;
            word = ((long) (fgetc(fp) & 0xff)) | (word << 0x08);
            word = ((long) (fgetc(fp) & 0xff)) | (word << 0x08);
            word = ((long) (fgetc(fp) & 0xff)) | (word << 0x08);
        }
        else {
            word =    (long)  fgetc(fp) & 0xff;
            word |= (((long) (fgetc(fp) & 0xff)) << 0x08);
            word |= (((long) (fgetc(fp) & 0xff)) << 0x10);
            word |= (((long) (fgetc(fp) & 0xff)) << 0x18);
        }
        return(word);
}

/*
**  Write a long word (32-bits) to a file stream.
**
**  Parameter values: byteorder : LSB_FIRST or MSB_FIRST
**                    fp        : FILE stream
**                    word      : 32-bit value to write
**  Returns: Nothing.
*/

void fputlong(char byteorder, FILE *fp, long word)
{
        if (byteorder == MSB_FIRST) {
                fputc((word >> 0x18) & 0xff, fp);
                fputc((word >> 0x10) & 0xff, fp);
                fputc((word >> 0x08) & 0xff, fp);
                fputc(word & 0xff, fp);
        }
        else {
                fputc(word & 0xff, fp);
                fputc((word >> 0x08) & 0xff, fp);
                fputc((word >> 0x10) & 0xff, fp);
                fputc((word >> 0x18) & 0xff, fp);
        }
}

/*
**  Read a short word (16-bits) from a file stream.
**
**  Parameter values: byteorder : LSB_FIRST or MSB_FIRST
**                    fp        : FILE stream
**
**  Returns: A 16-bit value in the specified format.
*/

short fgetshort(char byteorder, FILE *fp)
{
        register short word;

        if (byteorder == MSB_FIRST) {
                word = fgetc(fp) & 0xff;
                word = (fgetc(fp) & 0xff) | (word << 0x08);
        }
        else {
                word = fgetc(fp) & 0xff;
                word |= ((fgetc(fp) & 0xff) << 0x08);
        }
        return(word);
}

/*
**  Write a short word (16-bits) to a file stream.
**
**  Parameter values: byteorder : LSB_FIRST or MSB_FIRST
**                    fp        : FILE stream
**                    word      : 16-bit value to write
**  Returns: Nothing.
*/

void fputshort(char byteorder, FILE *fp, int word)
{
        if (byteorder == MSB_FIRST) {
                fputc((word >> 0x08) & 0xff, fp);
                fputc(word & 0xff, fp);
        }
        else {
                fputc(word & 0xff, fp);
                fputc((word >> 0x08) & 0xff, fp);
        }
}
