/***************************************************************************
* Program: FILES.C - An MS-DOS compatible file system
* Author: Philip J. Erdelsky
* Compiler: access_sectors() (and its page_violation() subroutine) and
*           get_date_and_time() are specific to Turbo C 2.0. When
*           compiling for other machines these need to be replaced.
*           Otherwise, the code should be usable with any compiler.
*           No system header files are required.
*
* Memory Models: any
* Compile-time switches:
*     STRUCTURE_ASSIGNMENT - if defined, structure assignment is used
*     REVERSE_BYTES        - if defined, use 68000 byte ordering
*     TURBOC               - if defined, compile for TURBOC
*     DRIVER               - if defined, a test driver is included
* Date: April 1990  May be used freely if authorship is acknowledged.
***************************************************************************/

/* This code is for Turbo C 2.0, it assumes the availability of biosdisk() */

#define TURBOC
#define DRIVER

/* This parameter is the same for all supported file systems. */

#define SECTOR_SIZE   512

/* Include this definition if C compiler permits structure assignment.*/

#define STRUCTURE_ASSIGNMENT

/* Include this definition if the CPU requires the MS byte of word to be */
/* at the lower address.                                                 */

#if 0
#define REVERSE_BYTES
#endif

/* Definitions of integer types used in file system. */

typedef unsigned char byte;  /* 8 bits  */
typedef unsigned int word;   /* 16 bits */
typedef unsigned long dword; /* 32 bits */

/* These parameters can be adjusted according to storage capacity and  */
/* system configuration.                                               */

#define NUMBER_OF_DRIVES              2
#define MAXIMUM_NUMBER_OF_FILES      16

/* The FAT cache must be large enough to accommodate the largest FAT  */
/* on any supported disk format:                                      */
/*   5-1/4 inch SSDD    2             3-1/2 inch SSDD    2            */
/*   5-1/4 inch DSDD    2             3-1/2 inch DSDD    ?            */
/*   5-1/4 inch DSHD    7             3-1/2 inch DSHD    9            */

#define MAXIMUM_FAT_SIZE              9

#include "files.h"

#define isalpha(x) ((('a'<=(x)) && ((x)<='z')) || (('A'<=(x)) && ((x)<='Z')))
#define toupper(x) ((('a'<=(x)) && ((x)<='z')) ? ('A'+(x)-'a') : (x))

#define NULL ((void *) 0)

/* Values of first argument in access_sectors() and gather(). */

#define  READ_SECTORS    0
#define  WRITE_SECTORS   1
#define  END_SERIES      2

/* Date and time information -- if the system does not have a real-time */
/* clock, use the default date and time 1/1/80 00:00:00.                */

struct date_and_time {
    int second; /* 0-59, default = 0 */
    int minute; /* 0-59, default = 0 */
    int hour;   /* 0-23, default = 0 */
    int day;    /* 1-31, default = 1 */
    int month;  /* 1-12, default = 1 */
    int year;   /* 1980-2107, default = 1980 */
};

/* Information maintained internally for diskette file system. */

struct fsys {
    int drive;
    word tracks;
    word sides;
    word sectors;
    word sectors_per_cluster;
    word first_FAT_sector;
    word sectors_per_FAT;
    word first_directory_sector;
    word first_data_sector;
    word last_cluster;
    word cluster_size;
    word open_files;
    byte buf[SECTOR_SIZE];
    byte FAT[SECTOR_SIZE*MAXIMUM_FAT_SIZE];
};

/* Directory structure as it appears on the diskette. */

struct directory {
    byte name[8];
    byte extension[3];
    byte attribute;
    byte reserved[10];
    word time;
    word date;
    word first_cluster;
    dword size;
};

#define ENTRIES_PER_SECTOR  (SECTOR_SIZE/sizeof(struct directory))

/* Special values for first byte of name. */

#define END_DIRECTORY 0x00
#define ERASED_FILE   0xE5

/* Bit values for attribute. */

#define READ_ONLY  0x01
#define HIDDEN     0x02
#define SYSTEM     0x04
#define VOLUME     0x08
#define DIRECTORY  0x10
#define ARCHIVE    0x20

/* Special values for FAT entries. */

#define EMPTY_CLUSTER     0x000
#define RESERVED_CLUSTER  0xFF0
#define BAD_CLUSTER       0xFF7
#define LAST_CLUSTER      0xFF8

/* Information maintained internally for open files. */

struct file {
    int mode;
    struct fsys *fsys;
    dword position;
    word cluster;
    word last_cluster;
    struct directory directory;
    word directory_sector;
    word directory_index;
};

/* Special mode bit indicating that file has been written to and should */
/* have its date and time updated and its ARCHIVE bit set when it is    */
/* closed.                                                              */

#define WRITTEN_TO   0x8000

/* Layout of first sector when read into fs->buf[].  A simple structure */
/* would not be portable because some words are not aligned on even     */
/* addresses.                                                           */

#define SECTORS_PER_CLUSTER  fs->buf[13]
#define RESERVED_SECTORS     (fs->buf[14]|fs->buf[15]<<8)
#define NUMBER_OF_FATS       fs->buf[16]
#define ROOT_ENTRIES         (fs->buf[17]|fs->buf[18]<<8)
#define TOTAL_SECTORS        (fs->buf[19]|fs->buf[20]<<8)
#define SECTORS_PER_FAT      (fs->buf[22]|fs->buf[23]<<8)
#define SECTORS_PER_TRACK    (fs->buf[24]|fs->buf[25]<<8)
#define HEADS                (fs->buf[26]|fs->buf[27]<<8)

/* Internal tables. */

static struct fsys fsys[NUMBER_OF_DRIVES];
static struct file file[MAXIMUM_NUMBER_OF_FILES];

/* Function prototypes */
static  int gather(int ,int ,int ,int ,int ,char *);
static  int access_sequential_sector(int ,struct fsys *, 
            unsigned int ,unsigned char *);
static  unsigned int FAT_entry(struct fsys *, unsigned int );
static  void set_FAT_entry(struct fsys *, unsigned int ,unsigned int);
static  void assign_directory(struct directory *, struct directory *);
static  int directory_lookup(struct file *, struct fsys *);
static  void set_date_and_time_in_directory(struct file *);
static  int invalid_file_character(int);
static  struct file *file_pointer(int);


#ifdef TURBOC
/* Drive interface for Turbo C 2.0 through IBM PC or compatible ROM BIOS. */

#include <bios.h>
#include <dos.h>

/* Function prototypes */
static  int page_violation(unsigned int,unsigned int, unsigned int);
static  int access_sectors(int, int, int, int, int, int, char *);
static  void get_date_and_time(struct date_and_time *);

/* This function returns a true (nonzero) value if a buffer of the 
 * specified length starting at segment:offset crosses a 64K page boundary */

static int page_violation(offset, segment, length)
unsigned offset, segment, length;
{
    long s = ((long) segment << 4)+offset;
    return((s & 0xF0000L) != ((s + length - 1) & 0xF0000L));
}

/* This function reads (or writes) the buffer contents from (or to) the   
 * specified sector or sectors, depending on the value of the func        
 * parameter (func=READ_SECTORS or func=WRITE_SECTORS). The side, track   
 * and sector parameters specify the first sector. The count is the number
 * of consecutive sectors. Higher-level code ensures that the operation   
 * will not extend past the end of the track.  */

static int access_sectors(func, drive, side, track, sector, count, buffer)
int func, drive, side, track, sector, count; char *buffer;
{
    int e, i;
    /* The following line protects the hardware from damage by an
       incorrect track number. It is appropriate for a 360K drive,
       and should be modified for other kinds of drives. */
    if (track < 0 || track > 79) return(-0x1FF);
    /* Some drives will generate errors when sectors 10 and 11 are
       accessed in a single BIOS call. */
    if (sector < 10 && sector + count > 10)
    {
        if ((e = access_sectors(func, drive, side, track, sector,
            10-sector, buffer))!=0) return e;
        return(access_sectors(func, drive, side, track, 10,
            count - 10 + sector, buffer + SECTOR_SIZE * (10 - sector)));
    }
    /* If the buffer crosses a 64K page boundary, the operation must
       be divided. */
    if (page_violation(FP_OFF((char far *) buffer),
        FP_SEG((char far *) buffer),
        count*SECTOR_SIZE))
    {
        static char buff[2*SECTOR_SIZE]; /* two temporary buffers */
        char *buf = buff; /* by default, use the first one */
        int partial_count = 0;
        do partial_count += 1;
        /* First perform as much of the operation as possible without
           crossing a 64K page boundary */
        while (!page_violation(FP_OFF((char far *) buffer),
            FP_SEG((char far *) buffer),
            partial_count*SECTOR_SIZE));
        if (partial_count>1 &&
            (e = access_sectors(func, drive, side, track, sector,
            partial_count-1, buffer))!=0)
            return(e);
        buffer += (partial_count-1)*SECTOR_SIZE;
        /* If the temporary buffer also crosses a 64K page boundary,
           use the adjacent temporary buffer. */
        if (page_violation(FP_OFF((char far *) buf),
            FP_SEG((char far *) buf),
            SECTOR_SIZE)) buf += SECTOR_SIZE;
        /* Use the temporary buffer to perform the part of the operation
           that crosses a 64K page boundary. */
        if (func == WRITE_SECTORS)
            for (i=0; i<SECTOR_SIZE; i++) buf[i] = buffer[i];
        e = access_sectors(func, drive, side, track, sector, 1, buf);
        if (e != 0) return(e);
        if (func == READ_SECTORS)
            for (i = 0; i < SECTOR_SIZE; i++) buffer[i] = buf[i];
        buffer += SECTOR_SIZE;
        count -= partial_count;
        /* Perform the rest of the operation. */
        if (count!=0)
            e = access_sectors(func, drive, side, track, 
                                sector, count, buffer);
        return(e);
    }
    /* Try reading three times; try writing only once. */
    for (i = func == READ_SECTORS ? 3 : 1; i>0; i--)
    {
        e = biosdisk(2+func, drive, side, track, sector, count, buffer);
        /* Code 0x11 (correctable error) is not considered an error because
           the data is correct. */
        if (e == 0 || e == 0x11) return(0);
        /* For other errors, reset the diskette system before retrying. */
        biosdisk(0, 0, 0, 0, 0, 0, NULL);
    }
    /* If there is still an error after retrying, reset diskette system
       and return error code, modified so it will not conflict with file
       system errors. */
    biosdisk(0, 0, 0, 0, 0, 0, NULL);
    return(-0x100-e);
}

/* Clock interface for Turbo C 2.0, through MS-DOS. */

static void get_date_and_time(dt) struct date_and_time *dt;
{
    struct date d;
    struct time t;
    int day;
    getdate(&d); day = d.da_day;
    gettime(&t);
    getdate(&d);
    /* If the date has changed (at midnight), get the time again. */
    if (day!=d.da_day) gettime(&t);
    dt->second = t.ti_sec;
    dt->minute = t.ti_min;
    dt->hour = t.ti_hour;
    dt->day = d.da_day;
    dt->month = d.da_mon;
    dt->year = d.da_year;
}
#endif

/* This function reads (or writes) the buffer contents from (or to) the   
 * specified single sector, depending on the value of the func parameter  
 * (func=READ_SECTORS or func=WRITE_SECTORS). It also gathers consecutive 
 * accesses of the same kind when the sectors are consecutive and on the  
 * same track and side and the buffers are consecutive in memory.  The    
 * special call gather(END_SERIES) is used force an actual read or write  
 * even though the next call on gather() might be for the next consecutive
 * sector. */                                                               

static int gather(func, drive, side, track, sector, buffer)
int func, drive, side, track, sector; char *buffer;
{
    struct pending {  /* pending operations that have been gathered */
        char *buffer;
        int func, side, track, sector;
        int count;
    };
    static struct pending pending[NUMBER_OF_DRIVES];
    struct pending *p = &pending[drive];
    /* First determine whether pending operations need to be performed
       because this one cannot be appended to them. */
    if (p->count!=0 &&
        (func == END_SERIES || side!=p->side ||
        func!=p->func || track!=p->track ||
        sector!=p->sector+p->count ||
        buffer!=p->buffer+SECTOR_SIZE*p->count))
    {
        int e;
        /* If so, call on access_sectors() to perform pending operations. */
        e = access_sectors(p->func, drive, p->side, p->track,
            p->sector, p->count, p->buffer);
        /* Then remove pending operations. */
        p->count = 0;
        if (e!=0) return(e);
    }
    if (func!=END_SERIES)
    {
        /* If there are no pending operations, make this one pending. */
        if (p->count == 0)
        {
            p->buffer = buffer;
            p->func = func;
            p->side = side;
            p->track = track;
            p->sector = sector;
            p->count = 1;
        }
        /* Otherwise append this operation to the pending ones. */
        else p->count += 1;
    }
    return(0);
}

/* This function converts a sequential sector number to a side, track
 * and sector number, according to the standard algorithm, and calls 
 * gather() to perform the operation specified by the f parameter    
 * (f=READ_SECTORS or f=WRITE_SECTORS).    */

static int access_sequential_sector(f, fs, n, buffer)
int f; struct fsys *fs; word n; byte *buffer;
{
    word sector;
    sector = n % fs->sectors + 1;  n /= fs->sectors;
    return(gather(f, fs->drive, n % fs->sides, n/fs->sides, sector, buffer));
}

/* This function gets FAT entry i from the FAT image in the specified file 
 * system.        */

static word FAT_entry(fs, i) struct fsys *fs; word i;
{
    byte *p;
    p = fs->FAT + i + (i >> 1);
    return((word) ((i & 1) ? p[1] << 4 | p[0] >> 4 : 
                    p[0] | p[1] << 8 & 0xF00));
}

/* This function sets FAT entry i to x in the FAT image in the specified
 * file system.   */                                                         

static void set_FAT_entry(fs, i, x) struct fsys *fs; word i, x;
{
    byte *p;
    p = fs->FAT + i + (i >> 1);
    if (i&1)
    {
        p[0] = p[0] & 0x0F | x << 4;
        p[1] = x >> 4;
    }
    else
    {
        p[0] = x;
        p[1] = p[1] & 0xF0 | x >> 8;
    }
}

/* This function moves a directory entry between a buffer read from (or
 * written to) disk and the file system information kept in memory.    
 * It reverses byte ordering if the byte order is different.  */

#ifdef REVERSE_BYTES
static void assign_directory(p, q)
struct special_directory
{
    byte name_extension_attribute_reserved[8+3+1+10];
    byte time[2];
    byte date[2];
    byte first_cluster[2];
    byte size[4];
} *p, *q;
{
    int i;
    for (i = 0; i < sizeof(p->name_extension_attribute_reserved; i++)
        p->name_attribute_extension_reserved[i] =
        q->name_attribute_extension_reserved[i];
    p->time[0] = q->time[1];
    p->time[1] = q->time[0];
    p->date[0] = q->date[1];
    p->date[1] = q->date[0];
    p->first_cluster[0] = q->first_cluster[1];
    p->first_cluster[1] = q->first_cluster[0];
    p->size[0] = q->size[3];
    p->size[1] = q->size[2];
    p->size[2] = q->size[1];
    p->size[3] = q->size[0];
}
#else
#ifdef STRUCTURE_ASSIGNMENT
static void assign_directory(p, q) struct directory *p, *q;
{
    *p = *q;
}
#else
static void assign_directory(p, q) char *p, *q;
{
    int n = sizeof(struct directory);
    while (n!=0)
    {
        *p++ = *q++; --n;
    }
}
#endif
#endif

/* This function looks up the name and extension in the root directory.   
 * If found, it puts the appropriate information into the file structure  
 * and returns 1. If not found, it returns 0. If there is a drive error,  
 * it returns the drive error code, which is neither 1 nor 0.  */

static int directory_lookup(f, fs) struct file *f; struct fsys *fs;
{
    word sector, index;
    for (sector=fs->first_directory_sector; sector<fs->first_data_sector; 
         sector++)
    {
        int e;
        if ((e = access_sequential_sector(READ_SECTORS, fs, sector, 
                 fs->buf)) != 0
            || (e = gather(END_SERIES, fs->drive, 0,0,0,0))!=0)
            return(e);
        for (index=0; index < SECTOR_SIZE; index += sizeof(struct directory))
        {
            int j;
            if (fs->buf[index] == END_DIRECTORY) return(0);
            for (j = 0; j < 11; j++)
                if (fs->buf[index+j] != f->directory.name[j]) break;
            if (j == 11)
            {
                f->directory_index = index;
                f->directory_sector = sector;
                assign_directory(&f->directory, 
                     (struct directory *) (fs->buf + index));
                return(1);
            }
        }
    }
    return(0);
}

/* This function sets the date and time in the directory image of an open
 * file to the value supplied by the system clock. If the system has no  
 * clock, substitute an arbitrary (but valid) date and time, such as     
 * 1/1/80 00:00:00. */                                                   

static void set_date_and_time_in_directory(f) struct file *f;
{
    struct date_and_time dt;
    get_date_and_time(&dt);
    f->directory.time = dt.second >> 1 | dt.minute << 5 | dt.hour << 11;
    f->directory.date = dt.day | dt.month << 5 | (dt.year-1980) << 9;
}

static int invalid_file_character(c) int c;
{
    return(c<' '  || c=='"' || c=='/' || c=='\\' || c=='[' || c==']' ||
        c==':' || c=='|' || c=='<' || c=='>'  || c=='+' || c=='=' ||
        c==';' || c==',');
}

/* This function opens a file and returns as its functional value either a 
 * file handle (which must be nonnegative), or an error code defined in
 * files.h (which must be negative).     */

int open(name, mode) char *name; int mode;
{
    struct file *f;
    struct fsys *fs;
    int h;
    int e;
    mode &= O_RDWR|O_CREAT|O_TRUNC|O_EXCL;
    /* Get a handle (file table entry). */
    for (h = 0; file[h].mode != 0; h++)
        if (h == MAXIMUM_NUMBER_OF_FILES-1) return(TOO_MANY_FILES);
    f = &file[h];
    /* Check the drive specification and set drive number (A=0, etc.). */
    {
        int d;
        if (!isalpha(name[0]) || name[1]!=':' ||
            (d=toupper(name[0])-'A')>=NUMBER_OF_DRIVES)
            return(INVALID_DRIVE_SPEC);
        f->fsys = fs = &fsys[d];
        fs->drive = d;
    }
    /* Edit the file name and extension into the directory entry. */
    {
        int i = 0;
        int c;
        name += 2;
        while ((c=(*name++))!=0 && c!='.')
        {
            if (i<8)
            {
                if (invalid_file_character(c)) return(INVALID_FILE_NAME);
                f->directory.name[i++] = toupper(c);
            }
        }
        while (i<8) f->directory.name[i++] = ' ';
        if (c=='.') while ((c=(*name++))!=0)
        {
            if (i<11)
            {
                if (invalid_file_character(c)) return(INVALID_FILE_NAME);
                f->directory.name[i++] = toupper(c);
            }
        }
        while (i<11) f->directory.name[i++] = ' ';
        if ((f->directory.name[0]&0xFF) == 0xE5) f->directory.name[0] = 0x05;
    }
    /* Exclude multiple accesses to same file, except multiple O_RDONLY. */
    {
        int i;
        for (i=0; i<MAXIMUM_NUMBER_OF_FILES; i++) 
        if (file[i].mode != 0)
        {
            if (file[i].mode&O_WRONLY || mode&O_WRONLY)
            {
                int j;
                for (j = 0; j < 11; j++)
                    if (f->directory.name[j] != file[i].directory.name[j]) 
                        break;
                if (j == 11) return(MULTIPLE_WRITE);
            }
        }
    }
    /* If there are no open files, read file parameters from first sector. */
    if (fs->open_files == 0)
    {
        word total_sectors;
        word i;
        if ((e = access_sectors(READ_SECTORS, fs->drive, 
                 0, 0, 1, 1, fs->buf))!=0)
            return(e);
        total_sectors = TOTAL_SECTORS;
        fs->sides = HEADS;
        fs->sectors = SECTORS_PER_TRACK;
        fs->tracks = total_sectors / (fs->sides * fs->sectors);
        fs->first_FAT_sector = RESERVED_SECTORS;
        fs->sectors_per_FAT = SECTORS_PER_FAT;
        fs->first_directory_sector =
            fs->first_FAT_sector + fs->sectors_per_FAT * NUMBER_OF_FATS;
        fs->sectors_per_cluster = SECTORS_PER_CLUSTER;
        fs->first_data_sector = fs->first_directory_sector +
            ROOT_ENTRIES / ENTRIES_PER_SECTOR;
        fs->last_cluster =
            (total_sectors-fs->first_data_sector) / SECTORS_PER_CLUSTER + 1;
        fs->cluster_size = SECTOR_SIZE * fs->sectors_per_cluster;
        for (i = 0; i < fs->sectors_per_FAT; i++)
             if ((e = access_sequential_sector(READ_SECTORS, fs,
                fs->first_FAT_sector+i, fs->FAT + i * SECTOR_SIZE)) != 0)
                return(e);
        if ((e = gather(END_SERIES, fs->drive, 0,0,0,0)) != 0) 
             return(e);
    }
    if ((e = directory_lookup(f,fs))<0) return(e);
    if (mode&O_RDONLY) /* RDONLY or RDWR modes */
    {
        if (e == 0 || f->directory.attribute&(VOLUME|DIRECTORY))
            return(FILE_NOT_FOUND);
        if (mode&O_WRONLY && f->directory.attribute&READ_ONLY)
            return(WRITE_ACCESS_DENIED);
    }
    else if (mode&O_WRONLY) /* WRONLY mode */
    {
        if (e!=0)
        {
            if (f->directory.attribute&(VOLUME|DIRECTORY))
                return(VOLUME_OR_DIRECTORY_NAME);
            if (mode&O_CREAT && mode&O_EXCL) return(FILE_ALREADY_EXISTS);
            if (f->directory.attribute & READ_ONLY) 
                return(WRITE_ACCESS_DENIED);
            if (mode&O_TRUNC)
            {
                word j;
                j = f->directory.first_cluster;
                while (EMPTY_CLUSTER<j && j<RESERVED_CLUSTER)
                {
                    word k;
                    k = FAT_entry(fs,j);
                    set_FAT_entry(fs,j,EMPTY_CLUSTER);
                    j = k;
                }
                f->directory.attribute = ARCHIVE;
                f->directory.first_cluster = EMPTY_CLUSTER;
                f->directory.size = 0L;
                mode |= WRITTEN_TO;
            }
        }
        else if (!(mode&O_CREAT)) return(FILE_NOT_FOUND);
        else
        {
            word index;
            word sector = fs->first_directory_sector;
            int j;
            while (1)
            {
                if ((e = access_sequential_sector(READ_SECTORS, fs, 
                         sector, fs->buf))!=0 ||
                    (e = gather(END_SERIES, fs->drive, 0,0,0,0)) != 0)
                    return(e);
                for (index = 0; index<SECTOR_SIZE; 
                     index += sizeof(struct directory))
                        if (fs->buf[index] == ERASED_FILE || 
                            fs->buf[index] == END_DIRECTORY)
                        break;
                if (index < SECTOR_SIZE) break;
                if (++sector == fs->first_data_sector) 
                    return(DIRECTORY_FULL);
            }
            f->directory.attribute = ARCHIVE;
            f->directory.first_cluster = EMPTY_CLUSTER;
            f->directory.size = 0L;
            for (j = 0; j < 10; j++) 
                 f->directory.reserved[j] = 0;
            f->directory_index = index;
            f->directory_sector = sector;
            set_date_and_time_in_directory(f);
            assign_directory((struct directory *) (fs->buf+index), 
                     &f->directory);
            if ((e = access_sequential_sector(WRITE_SECTORS, fs, sector, 
                     fs->buf)) != 0 ||
                (e = gather(END_SERIES, fs->drive, 0,0,0,0)) != 0)
                return(e);
            mode |= WRITTEN_TO;
        }
    }
    else return(FILE_MODE_ERROR);
    f->mode = mode;
    f->position = 0;
    f->last_cluster = EMPTY_CLUSTER;
    f->cluster = f->directory.first_cluster;
    fs->open_files += 1;
    return(h);
}

/* This function turns a handle returned by open() into a pointer to a 
 * structure containing information maintained internally about an open
 * file, or NULL if the handle does not refer to a currently open file. */

static struct file *file_pointer(handle) int handle;
{
    return(handle < 0 || handle >= MAXIMUM_NUMBER_OF_FILES ||
        file[handle].mode == 0 ? NULL : &file[handle]);
}

/* This function closes an open file, releasing the handle for re-use.      
 * If the file has been written to, it also updates the directory entry     
 * and writes the cached FAT to disk. It returns 0 if the operation is    
 * successful, or a nonzero error code otherwise. */

int close(handle) int handle;
{
    struct file *f;
    struct fsys *fs;
    if ((f=file_pointer(handle)) == NULL) return(INVALID_HANDLE);
    fs = f->fsys;
    if (f->mode&WRITTEN_TO)
    {
        int e;
        word i = fs->first_FAT_sector;
        while (i<fs->first_directory_sector)
        {
            word j;
            for (j = 0; j < fs->sectors_per_FAT; j++)
                if ((e = access_sequential_sector(WRITE_SECTORS, fs, i++,
                    fs->FAT + j * SECTOR_SIZE)) != 0) return(e);
        }
        f->directory.attribute |= ARCHIVE;
        set_date_and_time_in_directory(f);
        if ((e = access_sequential_sector(READ_SECTORS, fs,
            f->directory_sector, fs->buf)) != 0 ||
            (e = gather(END_SERIES, fs->drive, 0,0,0,0)) != 0)
            return(e);
        assign_directory((struct directory *)
            (fs->buf+f->directory_index), &f->directory);
        if ((e = access_sequential_sector(WRITE_SECTORS, fs,
            f->directory_sector, fs->buf))!=0 ||
            (e = gather(END_SERIES, fs->drive, 0,0,0,0))!=0)
            return(e);
    }
    f->mode = 0;
    fs->open_files -= 1;
    return(0);
}

/* This function reads the specifed number of bytes from an open file to
 * the specified memory buffer. It returns the number of bytes actually
 * read (which must necessarily be non-negative), or a negative error code
 * if an error was encountered.  */

int read(handle, buffer, count) int handle; int count; char *buffer;
{
    struct file *f;
    struct fsys *fs;
    dword size, position;
    int e = 0;
    int unread_bytes = count;
    if ((f=file_pointer(handle)) == NULL) return(INVALID_HANDLE);
    fs = f->fsys;
    size = f->directory.size;
    position = f->position;
    if ((f->mode&O_RDONLY) == 0) return(FILE_MODE_ERROR);
    while (unread_bytes > 0)
    {
        int n = unread_bytes;
        int rest_of_sector = SECTOR_SIZE - position % SECTOR_SIZE;
        word sector = fs->first_data_sector +
            (f->cluster - 2) * fs->sectors_per_cluster +
            (position/SECTOR_SIZE) % fs->sectors_per_cluster;
        if (size < position + n) n = size-position;
        if (n == 0) break;
        if (f->cluster > fs->last_cluster)
        {
            e = FILE_FORMAT_ERROR;
            break;
        }
        if (n>rest_of_sector) n = rest_of_sector;
        if (position % SECTOR_SIZE == 0 && n == SECTOR_SIZE)
        {
            /* read a whole sector */
            if ((e = access_sequential_sector(READ_SECTORS, 
                    fs, sector, buffer)) != 0)
                break;
        }
        else
        {
            /* read a partial sector */
            int x = position % SECTOR_SIZE;
            int i;
            if ((e = access_sequential_sector(READ_SECTORS,
                fs, sector, fs->buf))!=0 ||
                (e = gather(END_SERIES, fs->drive, 0,0,0,0))!=0)
                break;
            for (i = 0; i < n; i++) 
                buffer[i] = fs->buf[x+i];
        }
        buffer += n;
        position += n;
        unread_bytes -= n;
        if (position % (fs->cluster_size) == 0)
        {
            word next_cluster = FAT_entry(fs, f->cluster);
            if (next_cluster == LAST_CLUSTER) f->last_cluster = f->cluster;
            f->cluster = next_cluster;
        }
    }
    f->position = position;
    if (e == 0) e = gather(END_SERIES, fs->drive, 0,0,0,0);
    return(e ? e : count-unread_bytes);
}

/* This function writes the specifed number of bytes from the specified       
 * memory buffer to an open file. It returns the number of bytes actually 
 * written (which must necessarily be nonnegative), or a negative error    
 * if an error was encountered. If the number of bytes written was        
 * nonnegative but less than the specified number, there was insufficient  
 * disk space to write the rest. */

int write(handle, buffer, count) int handle, count; char *buffer;
{
    struct file *f;
    struct fsys *fs;
    dword size, position;
    int e = 0;
    int unwritten_bytes = count;
    if ((f=file_pointer(handle)) == NULL) return(INVALID_HANDLE);
    fs = f->fsys;
    size = f->directory.size;
    position = f->position;
    if ((f->mode&O_WRONLY) == 0) return(FILE_MODE_ERROR);
    while (unwritten_bytes>0)
    {
        word sector;
        int n = unwritten_bytes;
        int rest_of_sector = SECTOR_SIZE - position % SECTOR_SIZE;
        if (n>rest_of_sector) n = rest_of_sector;
        if (f->cluster == EMPTY_CLUSTER || f->cluster >= LAST_CLUSTER)
        {
            word new_cluster;
            for (new_cluster=2; new_cluster<=fs->last_cluster; new_cluster++)
                if (FAT_entry(fs, new_cluster) == EMPTY_CLUSTER) break;
            if (new_cluster>fs->last_cluster) break;
            if (f->last_cluster!=EMPTY_CLUSTER)
                set_FAT_entry(fs, f->last_cluster, new_cluster);
            f->cluster = f->last_cluster = new_cluster;
            set_FAT_entry(fs, new_cluster, LAST_CLUSTER);
            if (f->directory.first_cluster == EMPTY_CLUSTER)
                f->directory.first_cluster = new_cluster;
        }
        else if (f->cluster>fs->last_cluster)
        {
            e = FILE_FORMAT_ERROR;
            break;
        }
        sector = fs->first_data_sector +
            (f->cluster-2) * fs->sectors_per_cluster +
            (position/SECTOR_SIZE) % fs->sectors_per_cluster;
        if (position % SECTOR_SIZE == 0 && 
           ( n == SECTOR_SIZE || position + n >= size))
        {
            /* write a whole sector */
            if ((e = access_sequential_sector(WRITE_SECTORS, fs, sector, 
                 buffer))!=0)
                break;
        }
        else
        {
            /* write a partial sector */
            int x = position % SECTOR_SIZE;
            int i;
            if ((e = access_sequential_sector(READ_SECTORS, fs, sector, 
                 fs->buf))!=0
                || (e = gather(END_SERIES, fs->drive, 0,0,0,0)) != 0)
                break;
            for (i = 0; i < n; i++) fs->buf[x+i] = buffer[i];
            if ((e = access_sequential_sector(WRITE_SECTORS, fs, sector,
                 fs->buf))!=0)
                      break;
        }
        f->mode |= WRITTEN_TO;
        buffer += n;
        position += n;
        unwritten_bytes -= n;
        if (position>size) size = position;
        if (position%(fs->cluster_size) == 0)
        {
            word next_cluster = FAT_entry(fs, f->cluster);
            if (next_cluster == LAST_CLUSTER) f->last_cluster = f->cluster;
            f->cluster = next_cluster;
        }
    }
    f->position = position;
    f->directory.size = size;
    if (e == 0) e = gather(END_SERIES, fs->drive, 0,0,0,0);
    return(e ? e : count-unwritten_bytes);
}

/* This function seeks to the specified offset in an open file. Subsequent   
 * reads and writes will begin at the specified offset. It returns the       
 * resulting offset (relative to the beginning of the file), or a negative    
 * error code if there was an error.  */

long lseek(handle, offset, mode) int handle, mode; long offset;
{
    struct file *f;
    struct fsys *fs;
    dword position;
    if ((f=file_pointer(handle)) == NULL) return(INVALID_HANDLE);
    fs = f->fsys;
    position = f->position;
    switch (mode)
    {
        case SEEK_SET:
            position = offset;
            break;
        case SEEK_CUR:
            position += offset;
            break;
        case SEEK_END:
            position = f->directory.size+offset;
            break;
        default:
            return(LSEEK_MODE_ERROR);
    }
    if (position>f->directory.size)
        return(INVALID_LSEEK_OFFSET);
    {
        word cluster = f->directory.first_cluster;
        word i;
        for (i = position / fs->cluster_size; i > 0; i--)
        {
            word new_cluster = FAT_entry(fs, cluster);
            if (new_cluster == LAST_CLUSTER) f->last_cluster = cluster;
            cluster = new_cluster;
        }
        f->cluster = cluster;
    }
    f->position = position;
    return(position);
}

#ifdef DRIVER

#include <stdio.h>
#include <string.h>

#define LOCALBUF 127

#define FNAME (argv[ndx-1])
#define FMODE (argv[ndx])
int main(int argc, char **argv)
{
    int ndx;
    int handle, mode, error;
    char buffer[LOCALBUF+1];

    if (argc < 3)
    {
        printf("Usage is files fname mode [fname mode ...]\n");
        return;
    }

    for (ndx = 2; ndx < argc; ndx += 2)
    {
        printf("Processing %s, mode %s\n", FNAME, FMODE);
        mode = 0;
        if (strchr(FMODE, 'w')) mode |= O_WRONLY | O_CREAT | O_EXCL;
        if (strchr(FMODE, 'r')) mode |= O_RDONLY;
        if (mode == 0)
        {
            printf("  --mode is invalid\n\n");
            continue;
        }
        handle = Xopen(FNAME, mode);
        if (handle < 0)
        {
            printf("  --open() returned error code %d\n", handle);
            continue;
        }
        if (mode == O_RDONLY)
        { /* read the file */
            for (;;)
            {
                error = Xread(handle, buffer, LOCALBUF);
                if (error < 0)
                {
                    printf("  --read() returned error code %d\n", error);
                    break;
                }
                else if (error == 0)
                {
                    printf("  --end of file--\n");
                    break;
                }
                else
                {
                    buffer[error] = '\0';  /* ensure terminator */
                    printf("read %d bytes:\n%s<==end\n\n", error, buffer);
                }
            }
        }
        error = Xclose(handle);
        if (error < 0)
        {
            printf("  --close() returned error code %d\n", error);
            continue;
        }
    }
}
#endif